import torch.nn as nn
from aiia.model.Model import AIIAConfig, AIIABase
from transformers import PreTrainedModel
from .config import aiuNNConfig


class aiuNN(PreTrainedModel):
    config_class = aiuNNConfig

    def __init__(self, config: aiuNNConfig, base_model: PreTrainedModel = None):
        super().__init__(config)
        self.config = config

        # Build base layers without dynamic unpacking
        if base_model is not None:
            self.base_layers = self._copy_base_model_safely(base_model)
        else:
            self.base_layers = self._build_static_base_layers(config)

        # Bilinear upsampling head
        self.upsample = nn.Upsample(
            scale_factor=self.config.upsample_scale,
            mode='bilinear',
            align_corners=False
        )
        self.final_conv = nn.Conv2d(
            in_channels=self.config.hidden_size,
            out_channels=self.config.num_channels,
            kernel_size=3,
            padding=1
        )

    def _build_static_base_layers(self, config):
        """Build layers without dynamic unpacking to avoid CUDAGraph issues"""
        from collections import OrderedDict
        
        layers = OrderedDict()
        in_channels = config.num_channels
        
        for i in range(config.num_hidden_layers):
            layers[f'conv_{i}'] = nn.Conv2d(
                in_channels, config.hidden_size,
                kernel_size=config.kernel_size, padding=1
            )
            layers[f'activation_{i}'] = getattr(nn, config.activation_function)()
            layers[f'maxpool_{i}'] = nn.MaxPool2d(kernel_size=1, stride=1)
            in_channels = config.hidden_size
        
        return nn.Sequential(layers)

    def _copy_base_model_safely(self, base_model):
        """Copy base model layers without dynamic unpacking"""
        from collections import OrderedDict
        
        layers = OrderedDict()
        
        if hasattr(base_model, 'cnn'):
            # Copy layers individually instead of unpacking
            for i, layer in enumerate(base_model.cnn):
                layers[f'copied_layer_{i}'] = layer
        elif hasattr(base_model, 'shared_layer') and hasattr(base_model, 'unique_layers'):
            # Build explicitly instead of using extend and unpack
            layers['shared_layer'] = base_model.shared_layer
            layers['shared_activation'] = base_model.activation
            layers['shared_maxpool'] = base_model.max_pool
            
            for i, ul in enumerate(base_model.unique_layers):
                layers[f'unique_layer_{i}'] = ul
                layers[f'unique_activation_{i}'] = base_model.activation
                layers[f'unique_maxpool_{i}'] = base_model.max_pool
        
        return nn.Sequential(layers)

    def forward(self, x):
        x = self.base_layers(x)
        x = self.upsample(x)
        x = self.final_conv(x)
        return x


if __name__ == "__main__":
    from aiia import AIIABase, AIIAConfig
    # Create a configuration and build a base model.
    config = AIIAConfig()
    ai_config = aiuNNConfig()
    base_model = AIIABase(config)
    # Instantiate Upsampler from the base model (works correctly).
    upsampler = aiuNN(config=ai_config, base_model=base_model)
    # Save the model (both configuration and weights).
    upsampler.save_pretrained("aiunn")

    # Now load using the overridden load method; this will load the complete model.
    upsampler_loaded = aiuNN.from_pretrained("aiunn")
    print("Updated configuration:", upsampler_loaded.config.__dict__)
